#ifndef LEVEL_PATCH_PRECONDITIONER_HH
#define LEVEL_PATCH_PRECONDITIONER_HH

#include <string>

#include <dune/common/timer.hh>
#include <dune/common/fvector.hh>
#include <dune/common/bitsetvector.hh>

#include <dune/solvers/iterationsteps/lineariterationstep.hh>

#include "../../data-structures/levelcontactnetwork.hh"

#include "supportpatchfactory.hh"

#include <dune/localfunctions/lagrange/pqkfactory.hh>

#include <dune/fufem/boundarypatch.hh>
#include <dune/fufem/functiontools/boundarydofs.hh>


template <class PatchSolver, class MatrixType, class VectorType>
class LevelPatchPreconditioner : public LinearIterationStep<MatrixType, VectorType> {

public:
    enum Mode {additive, multiplicative};

private:
    using Base = LinearIterationStep<MatrixType, VectorType>;

    using PatchFactory = SupportPatchFactory<GridView, GridView>;
    using Patch = typename SupportPatchFactory::Patch;

    typedef typename BasisType::GridView GridView;
    typedef typename GridView::Grid GridType;

    static const int dim = GridType::dimension;

    const Mode mode_;

    const LevelContactNetwork<GridView, dim>& levelContactNetwork_;

    std::vector<Patch> patches_;
    PatchSolver patchSolver_;

    const GridType& grid_;
    const int level_;
    const BasisType basis_;

    size_t patchDepth_;

    MatrixType matrix_;
    std::vector<std::shared_ptr<OscLocalProblem<MatrixType, VectorType>> > localProblems_;

public:

    // for each coarse patch given by patchLevelBasis: set up local problem, compute local correction
    LevelPatchPreconditioner(const LevelContactNetwork<GridView, dim>& levelContactNetwork,
                             const BasisType& patchLevelBasis,
                             const LocalAssembler& localAssembler,
                             const std::vector<std::shared_ptr<LocalInterfaceAssembler>>& localInterfaceAssemblers,
                             const Mode mode = LevelPatchPreconditioner::Mode::additive) :
          levelInterfaceNetwork_(levelInterfaceNetwork),
          patchLevelBasis_(patchLevelBasis),
          localAssembler_(localAssembler),
          localInterfaceAssemblers_(localInterfaceAssemblers),
          mode_(mode),
          grid_(levelInterfaceNetwork_.grid()),
          level_(levelInterfaceNetwork_.level()),
          basis_(levelInterfaceNetwork_)
    {
        setPatchDepth();
        this->verbosity_ = QUIET;
    }

    // build support patches
    void build() {


        localProblems_.resize(vertexInElements.size());

        // init local fine level corrections
        if (this->verbosity_ == FULL) {
            std::cout << std::endl;
            std::cout << "---------------------------------------------" << std::endl;
            std::cout << "Initializing local fine grid corrections! Level: " << level_ << std::endl;

            Dune::Timer timer;
            timer.reset();
            timer.start();
        }

        const int patchLevel = patchLevelBasis_.faultNetwork().level();
        for (size_t i=0; i<vertexInElements.size(); i++) {
            if (this->verbosity_ != QUIET) {
                std::cout << i << std::endl;
                std::cout << "---------------" << std::endl;
            }

            Patch patchDofs;
            SupportPatchFactory<decltype(coarseGridView), decltype(fineGridView)> patchFactory(coarseGridView, fineGridView, vertexInElementsBody);
            patchFactory.build(seedObject, patchDofs, patchDepth_);

            if ((i+1) % 10 == 0 && this->verbosity_ == FULL) {
                std::cout << '\r' << std::floor((i+1.0)/patchLevelBasis_.size()*100) << " % done. Elapsed time: " << timer.elapsed() << " seconds. Predicted total setup time: " << timer.elapsed()/(i+1.0)*patchLevelBasis_.size() << " seconds." << std::flush;
            }
        }

        if (this->verbosity_ == FULL) {
            std::cout << std::endl;
            std::cout << "Total setup time: " << timer.elapsed() << " seconds." << std::endl;
            std::cout << "---------------------------------------------" << std::endl;
            timer.stop();
        }

        if (this->verbosity_ != REDUCED) {
            std::cout << "Level: " << level_ << " LevelPatchPreconditioner::build() - SUCCESS" << std::endl;
        }
    }

    void setPatchDepth(const size_t patchDepth = 0) {
        patchDepth_ = patchDepth;
    }

    void setVerbosity(VerbosityMode verbosity) {
        this->verbosity_ = verbosity;
    }

    virtual void setProblem(const MatrixType& mat, VectorType& x, const VectorType& rhs) {

      /*  VectorType rhs;
        rhs.resize(matrix_.N());
        rhs = 0;

        //print(localToGlobal, "localToGlobal: ");
        //print(boundaryDofs, "boundaryDofs: ");

        localProblems_[i] = std::make_shared<OscLocalProblem<MatrixType, VectorType>>(matrix_, rhs, localToGlobal, boundaryDofs);*/

        this->x_ = &x;
        this->rhs_ = &rhs;
        this->mat_ = Dune::stackobject_to_shared_ptr(mat);

        patchSolver_.setProblem(mat, x, rhs);
    }

    virtual void iterate() {
        if (mode_ == additive)
            iterateAdd();
        else
            iterateMult();
    }

    void iterateAdd() {
        *(this->x_) = 0;	

        VectorType x = *(this->x_);
        for (size_t i=0; i<patches_.size(); i++) {
            patchSolver_.setIgnore(patches_[i]);
            patchSolver_.solve();

            *(this->x_) += x;
        }
    }

    void iterateMult() {
        *(this->x_) = 0;

        VectorType x = *(this->x_);
        for (size_t i=0; i<patches_.size(); i++) {
            VectorType updatedRhs(*(this->rhs_));
            this->mat_->mmv(*(this->x_), updatedRhs);

            //step(i);
            //print(updatedRhs, "localRhs: ");
            //writeToVTK(basis_, updatedRhs, "/storage/mi/podlesny/data/faultnetworks/rhs/local", "exactvertexdata_step_"+std::to_string(i));

            patchSolver_.setProblem(*this->mat_, *(this->x_), updatedRhs);
            patchSolver_.setIgnore(patches_[i]);
            patchSolver_.solve();

            //print(it, "it: ");
            //print(x, "x: ");

            //writeToVTK(basis_, x, "/storage/mi/podlesny/data/faultnetworks/sol/local", "exactvertexdata_step_"+std::to_string(i));

            /*if (i==5) {
                writeToVTK(basis_, x, "/storage/mi/podlesny/data/faultnetworks/iterates/", "exactvertexdata_patchDepth_"+std::to_string(patchDepth_));
            }*/

            *(this->x_) += x;
        }
    }

    const GridView& gridView() const {
        return basis_.getGridView();
    }

    const LevelInterfaceNetwork<GridView>& levelInterfaceNetwork() const {
        return levelInterfaceNetwork_;
    }

    size_t size() const {
        return localProblems_.size();
    }

};

#endif

