#ifndef SUPPORT_PATCH_FACTORY_HH
#define SUPPORT_PATCH_FACTORY_HH

#include <set>
#include <queue>
#include <memory>

#include <dune/common/bitsetvector.hh>
#include <dune/common/fvector.hh>

#include <dune/fufem/referenceelementhelper.hh>
#include <dune/grid/common/mcmgmapper.hh>

#include <dune/localfunctions/lagrange/pqkfactory.hh>

//#include "../../data-structures/levelcontactnetwork.hh"
#include "hierarchicleveliterator.hh"

template <class Element>
class BodyElement {
private:
    size_t bodyID_;
    Element element_;

public:
    BodyElement() {}
    BodyElement(size_t bodyID, const Element& element) {
        set(bodyID, element);
    }

    void set(size_t bodyID, const Element& element) {
        bodyID_ = bodyID;
        element_ = element;
    }

    const auto& bodyID() const {
        return bodyID_;
    }
    const auto& element() const {
        return element_;
    }
};


template <class LevelContactNetwork>
class BodyHelper {
private:
    using GridType = typename LevelContactNetwork::DeformedGridType;
    static const int dim = GridType::dimension;

    using Element = typename GridType::template Codim<0>::Entity;
    //using Vertex = typename GridType::template Codim<dim>::Entity;

    using LevelGridView = typename LevelContactNetwork::DeformedLevelGridView;
    using LeafGridView = typename LevelContactNetwork::DeformedLeafGridView;

    using HierarchicLevelIteratorType = HierarchicLevelIterator<GridType>;
    using LevelElementMapper = typename Dune::MultipleCodimMultipleGeomTypeMapper<LevelGridView,  Dune::MCMGElementLayout >;
    using LevelFaceMapper = typename Dune::MultipleCodimMultipleGeomTypeMapper<LevelGridView,  mcmgLayout(Codim<1>()) >;

    using LeafElementMapper = typename Dune::MultipleCodimMultipleGeomTypeMapper<LeafGridView,  Dune::MCMGElementLayout >;
    using LeafFaceMapper = typename Dune::MultipleCodimMultipleGeomTypeMapper<LeafGridView, mcmgLayout(Codim<1>()) >;

    bool levelMode_;

    std::unique_ptr<LevelGridView> levelGridView_;
    LevelElementMapper levelElementMapper_;
    LevelFaceMapper levelFaceMapper_;

    std::unique_ptr<LeafGridView> leafGridView_;
    LeafElementMapper leafElementMapper_;
    LeafFaceMapper leafFaceMapper_;

    std::vector<std::vector<Element>>& vertexInElements_;

    //std::vector<size_t> globalToLocal_;

    template <class GV>
    void init(const GV& gv) {
        // init vertexInElements
        vertexInElements_.resize(gv.size(dim));
        for (size_t i=0; i<vertexInElements_.size(); i++) {
                vertexInElements_[i].resize(0);
        }

        const auto& indexSet = gv.indexSet();

        for (const auto& elem : elements(gv)) {
            const auto& lfe = cache_.get(elem.type());

            for (size_t j=0; j<lfe.localBasis().size(); ++j) {
                auto localIdx = lfe.localCoefficients().localKey(j).subEntity();
                int dofIndex = indexSet.subIndex(elem, localIdx, dim);

                vertexInElements_[dofIndex].push_back(elem);
            }
        }
    }
public:

    BodyHelper(const LevelGridView& gridView) :
        levelMode_(true),
        levelGridView_(std::make_unique<LevelGridView>(gridView)),
        levelElementMapper_(gridView),
        leafFaceMapper_(gridView) {

        init(gridView);
    }

    BodyHelper(const LeafGridView& gridView) :
        levelMode_(false),
        leafGridView_(std::make_unique<LeafGridView>(gridView)),
        leafElementMapper_(gridView),
        leafFaceMapper_(gridView) {

        init(gridView);
    }

    const auto& gridView() const {
        return (levelMode_ ? levelGridView_ : leafGridView_);
    }

    const auto& elementMapper() const {
        return (levelMode_ ? levelElementMapper_ : leafElementMapper_);
    }

    const auto& faceMapper() const {
        return (levelMode_ ? levelFaceMapper_ : leafFaceMapper_);
    }

    const auto& vertexInElements() const {
        return vertexInElements_;
    }

    const auto& cache() const {
        return cache_;
    }
};


/**
 *  constructs BitSetVector whose entries are
 *      true,   if vertex is outside of patch or part of patch boundary
 *      false,  if vertex is in interior of patch
 */
template <class LevelContactNetwork>
class SupportPatchFactory
{
public:
    using Patch = Dune::BitSetVector<1>;

private:
    static const int dim = LevelContactNetwork::dimension; //TODO
    using ctype = typename LevelContactNetwork::ctype;

    using BodyHelper = BodyHelper<LevelContactNetwork>;
    //using Vertex = typename GridType::template Codim<dim>::Entity;

    using Element = typename LevelContactNetwork::DeformedGridType::template Codim<0>::Entity;
    using BodyElement = BodyElement<Element>;
    using FiniteElementCache = Dune::PQkLocalFiniteElementCache<ctype, ctype, dim, 1>; // P1 finite element cache

    const size_t nBodies_;

    const LevelContactNetwork& coarseContactNetwork_;
    const LevelContactNetwork& fineContactNetwork_;

    std::vector<std::unique_ptr<BodyHelper>> bodyHelper_;

    std::map<size_t, std::vector<BodyElement>> neighborElements_; // map faceID to neighbor elements
    std::map<size_t, std::vector<size_t>> neighborElementDofs_; // map faceID to neighbor element dofs

    std::vector<size_t> vertexOffSet_;
    std::vector<size_t> faceOffSet_;
    std::vector<size_t> elementOffSet_;

    FiniteElementCache cache_;

public:
    SupportPatchFactory(const LevelContactNetwork& coarseContactNetwork, const LevelContactNetwork& fineContactNetwork) :
        nBodies_(coarseContactNetwork_.nBodies()),
        coarseContactNetwork_(coarseContactNetwork),
        fineContactNetwork_(fineContactNetwork),
        bodyHelper_(nBodies_),
        vertexOffSet_(nBodies_),
        faceOffSet_(nBodies_),
        elementOffSet_(nBodies_){

        assert(nBodies_ == fineContactNetwork_.nBodies());

        size_t vertexOffSet = 0;
        size_t faceOffSet = 0;
        size_t elementOffSet = 0;

        for (size_t i=0; i<nBodies_; i++) {
            const auto& gridView = coarseContactNetwork_.body(i)->gridView();
            bodyHelper_[i] = std::make_unique<BodyHelper>(gridView);

            vertexOffSet_[i] = vertexOffSet + gridView.size(dim);
            vertexOffSet = vertexOffSet_[i];

            faceOffSet_[i] = faceOffSet + gridView.size(1);
            faceOffSet = faceOffSet_[i];

            elementOffSet_[i] = elementOffSet + gridView.size(0);
            elementOffSet = elementOffSet_[i];
        }

        // init boundary mapping/transfer
        const auto& nBodyAssembler = coarseContactNetwork_.nBodyAssembler();
        const auto& contactCouplings = nBodyAssembler.getContactCouplings();

        for (size_t i=0; i<contactCouplings.size(); i++) {
            const auto& coupling = nBodyAssembler.getCoupling(i);
            const auto& contactCoupling = contactCouplings[i];

            const size_t nmBody = coupling.gridIdx_[0];
            const size_t mBody = coupling.gridIdx_[1];

            const auto& nmBoundary = contactCoupling.nonmortarBoundary();
            const auto& mBoundary = contactCoupling.mortarBoundary();

            const auto& nmIndexSet = nonmortarBoundary.gridView().indexSet();
            const auto& mIndexSet = mortarBoundary.gridView().indexSet();

            for (const auto& rIs : intersections(*contactCoupling.getGlue())) {
                const auto& inside = rIs.inside();
                const auto& outside = rIs.outside();

                int nmFaceIdx = nmIndexSet.subIndex(inside, rIs.indexInInside(), 1);
                int mFaceIdx = mIndexSet.subIndex(outside, rIs.indexInOutside(), 1);

                BodyElement nmBodyElement(nmBody, inside);
                BodyElement mBodyElement(mBody, outside);

                neighborElements_[nmFaceIdx].emplace_back(mBodyElement);
                neighborElements_[mFaceIdx].emplace_back(nmBodyElement);

                std::set<size_t> dofs;
                computeIntersectionDofs(nmIndexSet, rIs, dofs);
                neighborElementDofs_[nmFaceIdx].insert(neighborElementDofs_[nmFaceIdx].end(), dofs.begin(), dofs.end());

                computeIntersectionDofs(mIndexSet, rIs, dofs, false);
                neighborElementDofs_[mFaceIdx].insert(neighborElementDofs_[mFaceIdx].end(), dofs.begin(), dofs.end());
            }
        }

        // neighborElementDofs_ contains dofs of connected intersections that are neighbors to a face given by faceID,
        // boundary dofs are contained once, interior dofs multiple times
        // task: remove boundary dofs and redundant multiples of interior dofs
        for (auto& it : neighborElementDofs_) {
            auto& dofs = it.second;

            std::sort(dofs.begin(), dofs.end());
            auto& dof = dofs.begin();
            while (dof != dofs.end()) {
                auto next = dof;
                if (*dof != *(++next))
                    dof = dofs.erase(dof);
            }
            dofs.erase(std::unique(dofs.begin(), dofs.end()), dofs.end());
        }
    }

    void build(const size_t bodyID, const Element& coarseElement, const size_t localVertex, Patch& patchDofs, const int patchDepth = 0) {
        BodyElement seedElement(bodyID, coarseElement);

        // construct coarse patch
        std::vector<BodyElement> patchElements;
        patchElements.emplace_back(seedElement);

        std::set<size_t> visitedElements;
        visitedElements.insert(elementIndex(seedElement));

        std::set<size_t> patchVertices;
        patchVertices.insert(vertexIndex(bodyID, coarseElement, localVertex));

        std::queue<BodyElement> patchSeeds;
        patchSeeds.push(seedElement);

        for (size_t depth=0; depth<=patchDepth; depth++) {
            std::vector<BodyElement> nextSeeds;
            std::set<size_t> newPatchVertices;

            while (!patchSeeds.empty()) {
                const auto& patchSeed = patchSeeds.front();
                patchSeeds.pop();

                size_t elemIdx = elementIndex(patchSeed);

                if (visitedElements.count(elemIdx))
                    continue;

                visitedElements.insert(elemIdx);

                size_t bodyIdx = patchSeed.bodyID();
                const auto& elem = patchSeed.element();
                const auto& gridView = bodyHelper_[bodyIdx]->gridView();

                for (auto& it : intersections(gridView, elem)) {
                    if (isPatchIntersection(gridView.indexSet(), it, patchVertices, newPatchVertices)) {
                        if (it.neighbor()) {
                            BodyElement neighbor(bodyIdx, it.outside());

                            if (visitedElements.count(elementIndex(neighbor)))
                                continue;

                            patchElements.emplace_back(neighbor);
                            patchSeeds.push(neighbor);
                        } else {
                            size_t faceIdx = faceIndex(bodyIdx, it);

                            if (neighborElements_.count(faceIdx)) {
                                const auto& neighbors = neighborElements_[faceIdx];
                                for (size_t i=0; i<neighbors.size(); i++) {
                                    const auto& neighbor = neighbors[i];

                                    if (visitedElements.count(elementIndex(neighbor)))
                                        continue;

                                    patchVertices.insert(neighborElementDofs_[faceIdx].begin(), neighborElementDofs_[faceIdx].end());
                                    patchElements.emplace_back(neighbor);
                                    patchSeeds.push(neighbor);
                                }
                            }
                        }
                    } else {
                        if (it.neighbor()) {
                            BodyElement neighbor(bodyIdx, it.outside());

                            if (visitedElements.count(elementIndex(neighbor)))
                                continue;

                            nextSeeds.emplace_back(neighbor);
                        } else {
                            size_t faceIdx = faceIndex(bodyIdx, it);

                            if (neighborElements_.count(faceIdx)) {
                                const auto& neighbors = neighborElements_[faceIdx];
                                for (size_t i=0; i<neighbors.size(); i++) {
                                    const auto& neighbor = neighbors[i];

                                    if (visitedElements.count(elementIndex(neighbor)))
                                        continue;

                                    newPatchVertices.insert(neighborElementDofs_[faceIdx].begin(), neighborElementDofs_[faceIdx].end());
                                    nextSeeds.emplace_back(neighbor);
                                }
                            }
                        }
                    }
                }
            }

            for (size_t i=0; i<nextSeeds.size(); i++) {
                patchSeeds.push(nextSeeds[i]);
            }
            patchVertices.insert(newPatchVertices.begin(), newPatchVertices.end());
        }

        // construct fine patch
        for (size_t i=0; i<patchElements.size(); ++i) {
            addFinePatchElements(patchElements[i], visitedElements, patchDofs);
        }
    }

private:
    template <class IndexSet, class Intersection>
    bool isPatchIntersection(const IndexSet& indexSet, const Intersection& is, const std::set<size_t>& patchVertices, std::set<size_t>& newPatchVertices) {
        std::set<size_t> intersectionDofs;
        computeIntersectionDofs(indexSet, is, intersectionDofs);

        for (auto& dof : intersectionDofs) {
            if (patchVertices.count(dof)) {
                newPatchVertices.insert(intersectionDofs.begin(), intersectionDofs.end());
                return true;
            }
        }
        return false;
    }

    size_t vertexIndex(const size_t bodyID, const Element& elem, const size_t localVertex) const {
        const auto& gridView = bodyHelper_[bodyID]->gridView();

        auto localIdx = cache_.get(elem.type()).localCoefficients().localKey(localVertex).subEntity();
        return vertexOffSet_[bodyID] + gridView.indexSet().subIndex(elem, localIdx, dim);
    }

    template <class Intersection>
    size_t faceIndex(const size_t bodyID, const Intersection& is) const {
        const auto& gridView = bodyHelper_[bodyID]->gridView();
        return faceOffSet_[bodyID] + gridView.indexSet().subIndex(is.inside(), is.indexInInside(), 1);
    }

    size_t elementIndex(const BodyElement& bodyElem) const {
        const auto& gridView = bodyHelper_[bodyElem.bodyID()]->gridView();
        return elementOffSet_[bodyElem.bodyID()] + gridView.indexSet().index(bodyElem.element());
    }

    template <class Intersection>
    bool containsInsideSubentity(const Element& elem, const Intersection& intersection, int subEntity, int codim) {
	    return ReferenceElementHelper<double, dim>::subEntityContainsSubEntity(elem.type(), intersection.indexInInside(), 1, subEntity, codim);
	}
      
    void addFinePatchElements(const BodyElement& coarseElem, const std::set<size_t>& coarsePatch, Patch& patchDofs) {
        using HierarchicLevelIteratorType = HierarchicLevelIterator<typename LevelContactNetwork::DeformedGridType>;

        const auto& grid = coarseContactNetwork_.body(coarseElem.bodyID())->gridView().grid();
        const auto fineLevel = fineContactNetwork_.body(coarseElem.bodyID())->gridView().level();

        HierarchicLevelIteratorType endHierIt(grid, coarseElem.element(), HierarchicLevelIteratorType::PositionFlag::end, fineLevel);
        HierarchicLevelIteratorType hierIt(grid, coarseElem.element(), HierarchicLevelIteratorType::PositionFlag::begin, fineLevel);
        for (; hierIt!=endHierIt; ++hierIt) {
            const Element& fineElem = *hierIt;
            addLocalDofs(coarseElem, fineElem, coarsePatch, patchDofs);
	    }
	}	  
	
    void addLocalDofs(const BodyElement& coarseElem, const Element& fineElem, const std::set<size_t>& coarsePatch, Patch& patchDofs) {
	    // insert local dofs
        const auto& lfe = cache_.get(fineElem.type());
        const auto& gridView = fineContactNetwork_.body(coarseElem.bodyID())->gridView();
        const auto& indexSet = gridView.indexSet();

        for (size_t i=0; i<lfe.localBasis().size(); ++i) {
            auto localIdx = lfe.localCoefficients().localKey(i).subEntity();
            int dofIndex = indexSet.subIndex(fineElem, localIdx, dim);
            patchDofs[dofIndex][0] = false;
	    }
	    
        const auto coarseLevel = coarseElem.element().level();

	    // search for parts of the global and inner boundary 
        for (const auto& is : intersections(gridView, fineElem)) {
            bool isPatchBoundary = false;

            if (is.neighbor()) {
                auto outsideFather = is.outside();
		    
                while (outsideFather.level()>coarseLevel) {
                    outsideFather = outsideFather.father();
                }
		    
                BodyElement outside(coarseElem.bodyID(), outsideFather);

                if (!coarsePatch.count(elementIndex(outside))) {
                    isPatchBoundary = true;
                }
            } else {
                isPatchBoundary = true;

                auto outsideFather = is.outside();

                while (outsideFather.level()>coarseLevel) {
                    outsideFather = outsideFather.father();
                }

                BodyElement outside(coarseElem.bodyID(), outsideFather);

                size_t faceIdx = faceIndex(coarseElem.bodyID(), is);

                if (neighborElements_.count(faceIdx)) {
                    const auto& outside = neighbors[0];

                    if (coarsePatch.count(elementIndex(outside))) {
                        isPatchBoundary = false;
                    }
                }
            }
		
            if (isPatchBoundary) {
                const auto& localCoefficients = cache_.get(fineElem.type()).localCoefficients();

                for (size_t i=0; i<localCoefficients.size(); i++) {
                    auto entity = localCoefficients.localKey(i).subEntity();
                    auto codim  = localCoefficients.localKey(i).codim();

                    if (containsInsideSubentity(fineElem, is, entity, codim)) {
                        int dofIndex = indexSet.subIndex(fineElem, entity, dim);
                        patchDofs[dofIndex][0] = true;
                    }
                }
            }
	    }	
	}


    template <class IndexSet, class Intersection>
    void computeIntersectionDofs(const IndexSet& idxSet, const Intersection& intersection, std::set<size_t>& intersectionDofs, bool inside = true) {
        intersectionDofs.clear();

        Element* elem;
        size_t intersectionIndex;

        if (inside) {
            elem = &intersection.inside();
            intersectionIndex = intersection.indexInInside();
        } else {
            elem = &intersection.outside();
            intersectionIndex = intersection.indexInOutside();
        }

        const int dimElement = Element::dimension;
        const auto& refElement = Dune::ReferenceElements<double, dimElement>::general(elem->type());

        for (int i=0; i<refElement.size(intersectionIndex, 1, dimElement); i++) {
            size_t idxInElement = refElement.subEntity(intersectionIndex, 1, i, dimElement);
            size_t globalIdx = idxSet.subIndex(*elem, idxInElement, dimElement);

            intersectionDofs.insert(globalIdx);
        }
    }
};
#endif
