#ifndef DUNE_TNNMG_ITERATIONSTEPS_TRUNCATED_NONSMOOTH_NEWTON_MULTIGRID_HH
#define DUNE_TNNMG_ITERATIONSTEPS_TRUNCATED_NONSMOOTH_NEWTON_MULTIGRID_HH

#include <string>
#include <sstream>
#include <vector>
#include <iomanip>

#include <dune/common/timer.hh>

#include <dune/solvers/common/resize.hh>
#include "dune/solvers/iterationsteps/iterationstep.hh"
#include "dune/solvers/iterationsteps/lineariterationstep.hh"
#include <dune/solvers/solvers/iterativesolver.hh>
#include <dune/solvers/solvers/linearsolver.hh>
#include <dune/solvers/common/canignore.hh>

#include "localproblem.hh"


#include <dune/tnnmg/iterationsteps/linearcorrection.hh>

#include <dune/tectonic/../../src/utils/debugutils.hh>

namespace Dune {
namespace TNNMG {

/**
 * \brief One iteration of the TNNMG method
 *
 * \tparam F Functional to minimize
 * \tparam BV Bit-vector type for marking ignored components
 */
template<class F, class BV, class Linearization,
                                  class DefectProjection,
                                  class LineSearchSolver>
class TNNMGStep :
  public IterationStep<typename F::Vector, BV>
{
  using Base = IterationStep<typename F::Vector, BV>;

public:

  using Vector = typename F::Vector;
  using ConstrainedVector = typename Linearization::ConstrainedVector;
  using ConstrainedMatrix = typename Linearization::ConstrainedMatrix;
  using BitVector = typename Base::BitVector;
  using ConstrainedBitVector = typename Linearization::ConstrainedBitVector;
  using Functional = F;
  using IterativeSolver = Solvers::IterativeSolver< ConstrainedVector, Solvers::DefaultBitVector_t<ConstrainedVector> >;
  using LinearSolver = Solvers::LinearSolver< ConstrainedMatrix,  ConstrainedVector >;

  /** \brief Constructor with an iterative solver object for the linear correction
   * \param iterativeSolver This is a callback used to solve the constrained linearized system
   * \param projection This is a callback used to compute a projection into a defect-admissible set
   * \param lineSolver This is a callback used to minimize a directional restriction of the functional
   *        for computing a damping parameter
   */
  TNNMGStep(const Functional& f,
            Vector& x,
            std::shared_ptr<IterationStep<Vector,BitVector> > nonlinearSmoother,
            std::shared_ptr<IterativeSolver> iterativeSolver,
            const DefectProjection& projection,
            const LineSearchSolver& lineSolver)
  : Base(x),
    f_(&f),
    nonlinearSmoother_(nonlinearSmoother),
    //linearCorrection_(makeLinearCorrection<ConstrainedMatrix>(iterativeSolver)),
    projection_(projection),
    lineSolver_(lineSolver)
  {}

  /** \brief Constructor with a linear solver object for the linear correction
   * \param linearSolver This is a callback used to solve the constrained linearized system
   * \param projection This is a callback used to compute a projection into a defect-admissible set
   * \param lineSolver This is a callback used to minimize a directional restriction of the functional
   *        for computing a damping parameter
   */
  TNNMGStep(const Functional& f,
            Vector& x,
            std::shared_ptr<IterationStep<Vector,BitVector> > nonlinearSmoother,
            std::shared_ptr<LinearSolver> linearSolver,
            const DefectProjection& projection,
            const LineSearchSolver& lineSolver)
  : Base(x),
    f_(&f),
    nonlinearSmoother_(nonlinearSmoother),
    linearSolver_(linearSolver),
    projection_(projection),
    lineSolver_(lineSolver)
  {}

  /** \brief Constructor with a LinearIterationStep object for the linear correction
   * \param linearIterationStep This is a callback used to solve the constrained linearized system
   * \param projection This is a callback used to compute a projection into a defect-admissible set
   * \param lineSolver This is a callback used to minimize a directional restriction of the functional
   *        for computing a damping parameter
   */
  TNNMGStep(const Functional& f,
            Vector& x,
            std::shared_ptr<Solvers::IterationStep<Vector,BitVector> > nonlinearSmoother,
            std::shared_ptr<Solvers::LinearIterationStep<ConstrainedMatrix,ConstrainedVector> > linearIterationStep,
            unsigned int noOfLinearIterationSteps,
            const DefectProjection& projection,
            const LineSearchSolver& lineSolver)
  : Base(x),
    f_(&f),
    nonlinearSmoother_(nonlinearSmoother),
    //linearCorrection_(makeLinearCorrection(linearIterationStep, noOfLinearIterationSteps)),
    projection_(projection),
    lineSolver_(lineSolver)
  {}

  using Base::getIterate;

  void preprocess() override
  {
    nonlinearSmoother_->setIgnore(this->ignore());
    nonlinearSmoother_->preprocess();
  }

  void setPreSmoothingSteps(std::size_t i)
  {
    preSmoothingSteps_ = i;
  }

  /**
   * \brief Do one TNNMG step
   */
  void iterate() override
  {

    const auto& f = *f_;
    const auto& ignore = (*this->ignoreNodes_);
    auto& x = *getIterate();

    //std::cout << "TNNMGStep::iterate " << std::endl;

    //print(f.quadraticPart(), "f.quadraticPart():");

    //print(f.linearPart(), "f.linearPart():");

    // Nonlinear presmoothing
    for (std::size_t i=0; i<preSmoothingSteps_; ++i)
        nonlinearSmoother_->iterate();

    std::cout << "- nonlinear presmoothing: success" << std::endl;
    print(x, "TNNMG iterate after smoothing:");

    // Compute constraint/truncated linearization
    if (not linearization_)
      linearization_ = std::make_shared<Linearization>(f, ignore);

    linearization_->bind(x);

    auto&& A = linearization_->hessian();
    auto&& r = linearization_->negativeGradient();

    //print(A, "TNNMG Linear problem matrix:");
    //print(r, "TNNMG Linear problem rhs:");

    // Compute inexact solution of the linearized problem
    Solvers::resizeInitializeZero(correction_, x);
    Solvers::resizeInitializeZero(constrainedCorrection_, r);

  //  std::cout << "- computing linear correction..." << std::endl;

  /*  auto canIgnoreCast = std::dynamic_pointer_cast<CanIgnore<Dune::Solvers::DefaultBitVector_t<Vector>>>( linearSolver_ );
    if (canIgnoreCast)
      canIgnoreCast->setIgnore(*this->ignoreNodes_);*/

    LocalProblem<ConstrainedMatrix, ConstrainedVector> localProblem(A, r, *this->ignoreNodes_);
    Vector newR;
    localProblem.getLocalRhs(constrainedCorrection_, newR);

    /*print(*this->ignoreNodes_, "ignoreNodes:");
    print(A, "A:");
    print(localProblem.getMat(), "localMat:");*/

    linearSolver_->setProblem(localProblem.getMat(), constrainedCorrection_, newR);
    linearSolver_->preprocess();
    linearSolver_->solve();

   // linearCorrection_(A, constrainedCorrection_, r);

    std::cout << "- linear correction: success" << std::endl;

    //print(constrainedCorrection_, "contrained correction:");

    linearization_->extendCorrection(constrainedCorrection_, correction_);

    std::cout << "- extended correction: success" << std::endl;
    print(correction_, "correction:");

    // Project onto admissible set
    projection_(f, x, correction_);

    std::cout << "- projection onto admissible set: success" << std::endl;

    print(correction_, "correction:");

    double const correctionNorm = correction_.two_norm();
    correction_ /= correctionNorm;


    // Line search
    auto fv = directionalRestriction(f, x, correction_);

   /* std::cout << fv.quadraticPart() << " f.quadraticPart():"  << std::endl;
    std::cout << fv.linearPart() << " f.linearPart():" << std::endl;
    std::cout << fv.subDifferential(0) << " f.subDifferential(0):" << std::endl;
    std::cout << fv.domain()[0] << " " << fv.domain()[0] << " f.domain():" << std::endl;

    std::cout << "- setup directional restriction: success" << std::endl; */
    dampingFactor_ = 0;
    lineSolver_(dampingFactor_, fv, false);

   // std::cout << "- computing damping factor: success" << std::endl;
    if (std::isnan(dampingFactor_))
      dampingFactor_ = 0;
    correction_ *= dampingFactor_;

    x += correction_;
  }

  /**
   * \brief Export the last computed damping factor
   */
  double lastDampingFactor() const
  {
    return dampingFactor_;
  }

  /**
   * \brief Export the last used linearization
   */
  const Linearization& linearization() const
  {
    return *linearization_;
  }

private:

  const Functional* f_;

  std::shared_ptr<IterationStep<Vector,BitVector> > nonlinearSmoother_;
  std::size_t preSmoothingSteps_ = 1;

  std::shared_ptr<Linearization> linearization_;

  //! \brief linear correction
  std::shared_ptr<LinearSolver> linearSolver_;

  typename Linearization::ConstrainedVector constrainedCorrection_;
  Vector correction_;
  DefectProjection projection_;
  LineSearchSolver lineSolver_;
  double dampingFactor_;
};


} // end namespace TNNMG
} // end namespace Dune

#endif
